local AddonName, GGL = ...
--[[-----------------------------------------------------------------------------
Label Widget
Displays text and optionally an icon.
-------------------------------------------------------------------------------]]
local Type, Version = "GoldGoblinLootGUIWidget-ChallengeLine", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local ipairs, unpack = ipairs, unpack
local max, select, pairs = math.max, select, pairs
local tonumber = tonumber
local abs, floor = math.abs, math.floor

-- WoW APIs
local CreateFrame, UIParent = CreateFrame, UIParent

-- GGL
local GetMoneyTextShort = GGL.GetMoneyTextShort
local GetMoneyText = GGL.GetMoneyText

local TOP_ITEMS_LIST_SIZE = GGL.CHALLENGE_TT_MAX_TOP_ITEMS

--[[-----------------------------------------------------------------------------
Tooltip
-------------------------------------------------------------------------------]]
local ShowTooltipFrame, ToolTipFrameIsKeepOpen, GetToolTipFrame
do
    local frame
    local LoadedPlayerTemplate

    local MAIN_SOURCE_FLAG_ICONS = GGL.MAIN_SOURCE_FLAG_ICONS
    local SUB_SOURCE_FLAG_ICONS = GGL.SUB_SOURCE_FLAG_ICONS
    local MAIN_SOURCE_FLAG = GGL.MAIN_SOURCE_FLAG

    local ITEM_NUM_FORMAT = "%s x %d"

    local FRAME_FUNCTIONS = {
        ["KeepOpen"] = function(self, state)
            if state then
                self.content.displayName:SetPoint("RIGHT", self.closeButton, "LEFT", 2, 0)
                self.closeButton:Show()
            else
                self.content.displayName:SetPoint("RIGHT", self.content, 0, 0)
                self.closeButton:Hide()
            end
            self.keepOpen = state
        end,

        ["SetItem"] = function(self, buttonID, itemID, quantity, moneyValue, sourceItemID, mainSource, subSource)
            if self.keepOpen then return end
            local button = self.content.itemFrame.buttons[buttonID]
            if not button then return end

            if not itemID then
                button:Hide()
                return
            end

            local itemName, itemLink = GetItemInfo(itemID)

            button.icon:SetTexture(GetItemIcon(itemID))
            button.itemName:SetFormattedText(ITEM_NUM_FORMAT, itemLink or "", tonumber(quantity) or 0)
            button.itemValue:SetText(GetMoneyTextShort(moneyValue))

            -- Get source textures
            if sourceItemID and mainSource == MAIN_SOURCE_FLAG.ITEM then
                button.sourceIcon:SetTexture(GetItemIcon(sourceItemID))
                button.sourceIcon:Show()
            else
                mainSource = ( mainSource and MAIN_SOURCE_FLAG_ICONS[mainSource] ) and MAIN_SOURCE_FLAG_ICONS[mainSource] or nil
                subSource = ( subSource and SUB_SOURCE_FLAG_ICONS[subSource] ) and SUB_SOURCE_FLAG_ICONS[subSource] or nil
                if mainSource or subSource then
                    button.sourceIcon:SetTexture(subSource or mainSource)
                    button.sourceIcon:Show()
                else
                    button.sourceIcon:Hide()
                end
            end

            button.itemID = itemID

            button:Show()
        end,

        ["SetItemList"] = function(self, itemList)
            for i = 1, TOP_ITEMS_LIST_SIZE do
                if itemList[i] then
                    self:SetItem(i, unpack(itemList[i]))
                else
                    self:SetItem(i)
                end
            end
        end,
    }

    local function Frame_OnEnter(self)
        if not self.itemID then return end
        local tooltip = _G.GameTooltip
        tooltip:ClearLines()
        tooltip:SetOwner(UIParent,"ANCHOR_CURSOR")
        tooltip:SetItemByID(self.itemID)
        tooltip:Show()
    end

    local function Frame_OnLeave(self)
        _G.GameTooltip:Hide()
    end

    local function AddItemButton(parent)
        local frame = CreateFrame("Frame", nil, parent)
        frame:SetScript("OnEnter", Frame_OnEnter)
        frame:SetScript("OnLeave", Frame_OnLeave)
        frame:SetHeight(12)

        local icon = frame:CreateTexture(nil, "BACKGROUND")
        icon:SetPoint("LEFT", frame, "LEFT", 0, 0)
        icon:SetWidth(10)
        icon:SetHeight(10)
        frame.icon = icon

        local sourceIcon = frame:CreateTexture(nil, "BACKGROUND")
        sourceIcon:SetPoint("RIGHT", frame, 0, 0)
        sourceIcon:SetWidth(10)
        sourceIcon:SetHeight(10)
        frame.sourceIcon = sourceIcon

        local itemName = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
        itemName:SetPoint("LEFT", icon, "RIGHT", 1, 0)
        itemName:SetPoint("RIGHT", sourceIcon, "LEFT", -1, 0)
        itemName:SetJustifyH("LEFT")
        itemName:SetJustifyV("MIDDLE")
        itemName:SetWordWrap(false)
        itemName:SetHeight(10)
        frame.itemName = itemName

        local itemValue = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
        itemValue:SetAllPoints(itemName)
        itemValue:SetJustifyH("RIGHT")
        itemValue:SetJustifyV("MIDDLE")
        itemValue:SetWordWrap(false)
        frame.itemValue = itemValue

        return frame
    end

    local function OnDisplayNameUpdate(playerTemplate, displayName)
        frame.content.displayName:SetText(playerTemplate.displayName)
    end

    local function GetFrame()
        if not frame then
            frame = CreateFrame("Frame", nil, UIParent, "TooltipBorderedFrameTemplate")
            frame:SetWidth(300)

            local closeButton = CreateFrame("Button", nil, frame, "UIPanelCloseButton")
            closeButton:SetPoint("TOPRIGHT", 2, 2)
            closeButton:SetScript("OnClick", function(self)
                frame:Hide()
                frame:KeepOpen(false)
            end)
            closeButton:Hide()
            frame.closeButton = closeButton

            -- ## content
            local content = CreateFrame("Frame", nil, frame)
            content:SetPoint("TOPLEFT", frame, 7, -7)
            content:SetPoint("BOTTOMRIGHT", frame, -7, 7)
            frame.content = content

            local factionIcon = content:CreateTexture(nil, "BACKGROUND")
            factionIcon:SetPoint("TOPLEFT", content, 0, 0)
            factionIcon:SetWidth(14)
            factionIcon:SetHeight(14)
            content.factionIcon = factionIcon

            local displayName = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
            displayName:SetPoint("LEFT", factionIcon, "RIGHT", 1, 0)
            displayName:SetPoint("RIGHT", content, 0, 0)
            displayName:SetJustifyH("LEFT")
            displayName:SetJustifyV("MIDDLE")
            displayName:SetHeight(14)
            content.displayName = displayName

            local battleTag = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
            battleTag:SetAllPoints(displayName)
            battleTag:SetJustifyH("RIGHT")
            battleTag:SetJustifyV("MIDDLE")
            battleTag:SetTextColor(0, 0.44, 0.87)
            content.battleTag = battleTag

            local lineName = content:CreateTexture(nil, "BACKGROUND")
            lineName:SetColorTexture(1, 1, 1, 0.5)
            lineName:SetHeight(1)
            lineName:SetPoint("TOPLEFT", factionIcon, "BOTTOMLEFT", -1, -2)
            lineName:SetPoint("RIGHT", content, 1, 0)


            -- itemList
            local itemFrame = CreateFrame("Frame", nil, content)
            itemFrame:SetPoint("TOPLEFT", factionIcon, "BOTTOMLEFT", 0, -5)
            itemFrame:SetPoint("RIGHT", content, 0, 0)
            content.itemFrame = itemFrame

            -- create items
            itemFrame.buttons = {}
            for i = 1, TOP_ITEMS_LIST_SIZE do
                local button = AddItemButton(itemFrame)

                if i == 1 then
                    button:SetPoint("TOPLEFT", itemFrame, 0, 0)
                else
                    button:SetPoint("TOPLEFT", itemFrame.buttons[i-1], "BOTTOMLEFT", 0, 0)
                end
                button:SetPoint("RIGHT", itemFrame, 0, 0)

                itemFrame.buttons[i] = button
            end
            itemFrame:SetHeight(TOP_ITEMS_LIST_SIZE * 12)



            frame:SetHeight(14 + 19 + itemFrame:GetHeight())

            -- setup methods
            for k,v in pairs(FRAME_FUNCTIONS) do
                frame[k] = v
            end
            FRAME_FUNCTIONS = nil
        end
        return frame
    end

    function ShowTooltipFrame(parent, playerTemplate)
        local frame = frame or GetFrame()
        local content = frame.content

        if LoadedPlayerTemplate then
            LoadedPlayerTemplate:UnregisterCallbackEvent("DisplayNameUpdate", OnDisplayNameUpdate)

            LoadedPlayerTemplate = nil
        end

        -- set name
        content.displayName:SetText(playerTemplate.displayName)

        -- set facionIcon
        if playerTemplate.factionName == "Horde" then
            content.factionIcon:SetAtlas("communities-create-button-wow-horde")
        elseif playerTemplate.factionName == "Alliance" then
            content.factionIcon:SetAtlas("communities-create-button-wow-alliance")
        elseif playerTemplate.factionName == "Neutral" then
            content.factionIcon:SetAtlas("AlliedRace-Crest-Nightborne")
        else
            content.factionIcon:Hide()
        end

        -- set battleTag
        if playerTemplate:GetType() == "BNWoWPlayer" then
			content.battleTag:SetText(playerTemplate.bnPlayer.battleTag)
        elseif playerTemplate:GetType() == "WoWPlayer" then
            content.battleTag:SetText("")
        end


        -- register callbacks
        playerTemplate:RegisterCallbackEvent("DisplayNameUpdate", OnDisplayNameUpdate)

        LoadedPlayerTemplate = playerTemplate

        frame.playerTemplate = playerTemplate

        frame:SetParent(parent)
        frame:SetPoint("CENTER")
        frame:Show()

        return frame
    end

    function ToolTipFrameIsKeepOpen()
        if frame and frame.keepOpen then
            return true
        else
            return false
        end
    end

    function GetToolTipFrame()
        return frame
    end
end

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]
local INVITE_KEY_LIST = GGL.INVITE_KEY_LIST
local INVITE_KEY_LIST_ATLAS_ICONS = GGL.INVITE_KEY_LIST_ATLAS_ICONS

local GAP_SIZE = 2

local FONT_SIZES = {
    [1] = _G.GameFontHighlightLarge,
    [2] = _G.GameFontHighlight,
    [3] = _G.GameFontHighlightSmall,
}
local GOLD_TEXT, SILVER_TEXT, COPPER_TEXT = "|cffffd70ag|r", "|cffc7c7cfs|r", "|cffeda55fc|r"

local function SetMoneyText(self, money)
    self:SetText(GetMoneyTextShort(money))
end

local function UpdatePoints(self)
    if self.noUpdate then return end
    local frame = self.frame
    local rank = self.rank
    local height = self.name:GetStringHeight() or 20

    -- Set points
    self.image:SetWidth((height*0.5)+4)
    self.image:SetHeight((height+4))

    rank:SetWidth( ( rank:GetUnboundedStringWidth() > (height+10) ) and ( rank:GetUnboundedStringWidth() ) or ( height+10 ) )
    rank:SetHeight(height)

    self.name:SetPoint("RIGHT", frame, "RIGHT", -(self.value:GetStringWidth()+2), 0)

    -- Set new frame height
    self.resizing = true
	frame:SetHeight(height + 4 + 1)
	frame.height = height + 4
	self.resizing = nil
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
    ["OnAcquire"] = function(self)
        self.sortValue = nil
        self.noUpdate = nil
        self:SetMoney(0, false, true)
        self:SetFontSize(3)
        self:ResetToolTipPoint()
        self:SetInviteState()
    end,

    ["OnRelease"] = function(self)
        self:SetPlayer(nil)
        self.frame:Hide()
    end,

    ["OnWidthSet"] = function(self, width)
        UpdatePoints(self)
    end,

    ["SetToolTipPoint"] = function(self, ttPoint, parent)
        self.ttPoint = ttPoint or "BOTTOM_X"
        self.ttParent = parent or self.frame
    end,

    ["ResetToolTipPoint"] = function(self)
        self.ttPoint = "BOTTOM_X"
        self.ttParent = self.frame
    end,

    ["SetPlayer"] = function(self, player, itemList)
        if self.player and self.player == player then return end
        if self.player then
            self.player:UnregisterCallbackEvent("DisplayNameUpdate", self.__onDisplayNameUpdate)
            self.player = nil
            self.itemList = nil
        end
        if player then
            self:SetName(player.displayName)
            --self:SetFaction(player.factionName)
            player:RegisterCallbackEvent("DisplayNameUpdate", self.__onDisplayNameUpdate)
        end
        self.itemList = itemList
        self.player = player
    end,

    --- Unregisters all events that updates the player
    -- If the update is needed again a new SetPlayer call is needed
    ["UnregisterPlayerUpdates"] = function(self)
        if self.player then
            self.player:UnregisterCallbackEvent("DisplayNameUpdate", self.__onDisplayNameUpdate)
        end
    end,

    ["GetPlayer"] = function(self)
        return self.player
    end,

    ["SetRank"] = function(self, rank, sizeUpdate)
        if not rank then return end
        local lastRank = self.rankValue

        if lastRank and lastRank ~= rank and not self.inviteState then
            self.image:Show()
            local top, bottom
            local rankDif
            if lastRank > rank then
                -- up
                top = 0
                bottom = 0.5
                rankDif = lastRank - rank - 1
            elseif lastRank < rank then
                -- down
                top = 0.5
                bottom = 1
                rankDif = rank - lastRank - 1
            end
            rankDif = rankDif > 3 and 3 or rankDif
            self.image:SetTexCoord(
                rankDif*0.25,
                (rankDif*0.25)+0.25,
                top,
                bottom
            )
        elseif not self.inviteState then
            self.image:Hide()
        end

        if sizeUpdate then
            self:SetFontSize(rank >= 3 and 3 or rank)
        end

        self.rankValue = rank
        self.rank:SetText(rank..".")
    end,

    ["GetRank"] = function(self)
        return self.rankValue or 0
    end,

    ["SetOffline"] = function(self, state)

    end,

    ["SetInviteState"] = function(self, state)
        if state == INVITE_KEY_LIST.INVITED then
            self.inviteState = INVITE_KEY_LIST.INVITED
            self.image:SetAtlas(INVITE_KEY_LIST_ATLAS_ICONS[INVITE_KEY_LIST.INVITED])
            self.image:SetTexCoord(0,1,0,1)
            self.image:Show()
        elseif state == INVITE_KEY_LIST.DECLINED then
            self.inviteState = INVITE_KEY_LIST.DECLINED
            self.image:SetAtlas(INVITE_KEY_LIST_ATLAS_ICONS[INVITE_KEY_LIST.DECLINED])
            self.image:SetTexCoord(0,1,0,1)
            self.image:Show()
        elseif self.inviteState then
            -- when we have a inviteState just refresh rank
            local rank = self.rankValue
            self.inviteState = nil
            self.image:SetTexture(GGL.MEDIA_PATH.."RankArrows")
            self.rankValue = nil
            self:SetRank(rank, true)
        end
    end,

    ["SetName"] = function(self, name)
        self.name:SetText(name)
    end,

    ["GetName"] = function(self, name)
        return self.name:GetText()
    end,

    ["SetMoney"] = function(self, copper, add, instantUpdate)
        self.money = add and ((self.money or 0) + copper) or (copper or 0)
        if instantUpdate then
            self.value:SetMoneyText(self.money)
            UpdatePoints(self)
        end
        self:UpdateSortValue()
    end,

    ["GetMoney"] = function(self)
        return self.money or 0
    end,

    ["UpdateSortValue"] = function(self)
        self:SetSortValue(self.money)
    end,

    ["SetSortValue"] = function(self, value)
        self.sortValue = value or 0
    end,

    ["GetSortValue"] = function(self)
        return self.sortValue or 0
    end,

    ["UpdateListRank"] = function(self, newRank, listUpdate)
        self.noUpdate = true
            self.value:SetMoneyText(self.money)
            self:SetRank(newRank, true)
            if newRank % 2 == 0 then
                self.bg:Show()
            else
                self.bg:Hide()
            end
        self.noUpdate = false
        UpdatePoints(self)
    end,

    ["SetFontSize"] = function(self, size)
        size = size or 3
        if self.fontSize ~= size and FONT_SIZES[size] then
            self.fontSize = size
            self:SetFontObjectRaw(FONT_SIZES[size])
            UpdatePoints(self)
        end
    end,

	["SetFontRaw"] = function(self, font, height, flags)
        self.rank:SetFont(font, height, flags)
        self.name:SetFont(font, height, flags)
        self.value:SetFont(font, height, flags)
	end,

	["SetFontObjectRaw"] = function(self, font)
        self.rank:SetFont((font or GameFontHighlightSmall):GetFont()) --GameFontNormal
        self.name:SetFont((font or GameFontHighlightSmall):GetFont())
        self.value:SetFont((font or GameFontHighlightSmall):GetFont())
    end,

    ["MergeInto"] = function(self, targetLine)
        if not targetLine then return end

        targetLine:SetMoney(self:GetMoney(), true)
    end,
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function Frame_OnEnter(self)
    if ToolTipFrameIsKeepOpen() or not self.obj.itemList or not self.obj.ttParent then return end

    local ttFrame = ShowTooltipFrame(self.obj.ttParent, self.obj.player)
    ttFrame:SetItemList(self.obj.itemList)
    GGL.GUI:SetToolTipPoint_Frame(ttFrame, self.obj.ttParent, "BOTTOM_X")

    self.ttFrame = ttFrame
end

local function Frame_OnLeave(self)
    if not self.ttFrame then return end
    if not ToolTipFrameIsKeepOpen() then
        self.ttFrame:Hide()
    end

    self.ttFrame = nil
end

local function Frame_OnClick(self, button)
    if not self.obj.itemList or not self.obj.ttParent then return end

    local ttFrame = GetToolTipFrame()
    if ToolTipFrameIsKeepOpen() then
        if self.obj.player == ttFrame.playerTemplate then
            ttFrame:KeepOpen(false)
            Frame_OnEnter(self)
        else
            ttFrame:KeepOpen(false)
            Frame_OnEnter(self)
            ttFrame:KeepOpen(true)
        end
    else
        ttFrame:KeepOpen(true)
    end
end

local function Constructor()
    local widget = {}
    widget.type  = Type

    local frame = CreateFrame("BUTTON", nil, UIParent)
    frame.obj = widget
    widget.frame = frame
    frame:SetScript("OnEnter", Frame_OnEnter)
    frame:SetScript("OnLeave", Frame_OnLeave)
    frame:SetScript("OnClick", Frame_OnClick)

    local bg = frame:CreateTexture(nil, "BACKGROUND", nil, 1)
    bg:SetAllPoints(frame)
    bg:SetColorTexture(1,1,1,GGL.LINE_ALPHA_VALUE)
    bg:Hide()
    widget.bg = bg

    local image = frame:CreateTexture(nil, "BACKGROUND", nil, 2)
    image:SetPoint("LEFT", frame, "LEFT", 1, 0)
    image:SetTexture(GGL.MEDIA_PATH.."RankArrows")
    image:Hide()
    widget.image = image

    local rank = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    rank:SetPoint("LEFT", image, "RIGHT", 0, 0)
    rank:SetJustifyH("CENTER")
    rank:SetJustifyV("RIGHT")
    rank:SetWordWrap(false)
    rank:SetNonSpaceWrap(false)
    widget.rank = rank

    local name = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    name:SetPoint("TOP", frame, "TOP", 0, 0)
    name:SetPoint("LEFT", rank, "RIGHT", GAP_SIZE, 0)
    name:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -1, 0)
    name:SetJustifyH("LEFT")
    name:SetJustifyV("MIDDLE")
    name:SetWordWrap(false)
    widget.name = name

    local value = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    value:SetPoint("TOPLEFT", rank, "TOPRIGHT", GAP_SIZE, 0)
    value:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -1, 0)
    value:SetJustifyH("RIGHT")
    value:SetJustifyV("MIDDLE")
    value:SetWordWrap(false)
    value.SetMoneyText = SetMoneyText
    widget.value = value

    -- methodes
    for method, func in pairs(methods) do
        widget[method] = func
    end

    -- cb
    widget.__onDisplayNameUpdate = function(self, displayName)
        widget:SetName(displayName)
    end

    -- init
    widget.fontSize = 3

    return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)
