local AddonName, GGL = ...
--[[-----------------------------------------------------------------------------
Label Widget
Displays text and optionally an icon.
-------------------------------------------------------------------------------]]
local Type, Version = "GoldGoblinLootGUIWidget-Timer", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local max, min, mod = math.max, math.min, mod
local select, pairs = select, pairs
local tonumber = tonumber
local abs, floor = math.abs, math.floor
local format = string.format
local C_TimerAfter = C_Timer.After

-- WoW APIs
local STOPWATCH_TIME_UNIT = _G.STOPWATCH_TIME_UNIT

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]
local SEC_TO_MINUTE_FACTOR = 1/60;
local SEC_TO_HOUR_FACTOR = SEC_TO_MINUTE_FACTOR*SEC_TO_MINUTE_FACTOR;
local TIMER_INTERVAL = 1

local function OnAcquire(self)

end

local function OnRelease(self)
    self:StopTimer(self)
end

local function Timer_SetData(self)
    local timer
    if self.isCounter then
        timer = self.startTime and self.timeFunc() - self.startTime or 0
    else
        timer = self.endTime and self.endTime - self.timeFunc() or 0
    end

    --local hour = min(floor(timer*SEC_TO_HOUR_FACTOR), 99)
    --local minute = mod(timer*SEC_TO_MINUTE_FACTOR, 60)
    --local second = mod(timer, 60)
    self.timer_lH:SetFormattedText(STOPWATCH_TIME_UNIT, min(floor(timer*SEC_TO_HOUR_FACTOR), 99))
    self.timer_lM:SetFormattedText(STOPWATCH_TIME_UNIT, mod(timer*SEC_TO_MINUTE_FACTOR, 60))
    self.timer_lS:SetFormattedText(STOPWATCH_TIME_UNIT, mod(timer, 60))
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {

    ["OnAcquire"] = function(self)
        self:SetSize()
        self:SetTimeFunction()
	end,

    ["OnRelease"] = function(self)
        self:StopTimer()
    end,

    ["IsCounter"] = function(self, flag)
        self.isCounter = flag
    end,

    ["StartTimer"] = function(self)
        if not self.duration then return end
        if self.isPause or not self.endTime then
            self.endTime = self.timeFunc() + self.duration
            self.isPause = false
        end
        self.isRunning = true
        self.startTime = self.timeFunc()
        C_TimerAfter(TIMER_INTERVAL, self.TimerUpdate)
    end,

    ["SetTimeFunction"] = function(self, func)
        self.timeFunc = func or GGL.GetTime
    end,

    ["PauseTimer"] = function(self)
        if not self.endTime then return end
        self.isRunning = false
        self.isPause = true
        self.duration = self.endTime - self.timeFunc()
    end,

    ["StopTimer"] = function(self)
        self.isRunning = false
        self.isPause = false
        self.duration = nil
        self.endTime = nil
        self.startTime = nil
        self:SetDuration(0)
    end,

    ["FinishTimer"] = function(self)
        if self.isRunning then
            self:StopTimer(self)
            self:Fire("OnTimerFinish")
        end
    end,

    ["SetDuration"] = function(self, duration)
        self.duration = duration
        Timer_SetData(self)
        if self.isRunning then
            if not duration or (duration and duration > 0) then
                self:StartTimer()
            else
                self:StopTimer()
            end
        end
    end,

    ["SetTime"] = function(self, startTime, endTime)
        if not startTime or not endTime then return end
        self:StopTimer()
        self.endTime = endTime
        self.startTime = startTime
        self.duration = endTime - ( startTime or self.timeFunc() )
        self.isRunning = false
        Timer_SetData(self)
    end,

    ["SetSize"] = function(self, size)
        local fontName, fontHeight, fontFlags
        local size1, size2
        local fHeight
        if not size or size == "large" then
            fontName, fontHeight, fontFlags = GameFontHighlightLarge:GetFont()
            size1, size2 = 23, 6
            fHeight = 20
        elseif size == "small" then
            fontName, fontHeight, fontFlags = GameFontHighlightSmall:GetFont()
            size1, size2 = 15, 6
            fHeight = 12
        else
            return self:SetSize()
        end

        self.frame:SetWidth(3*size1+2*size2)
        self.frame:SetHeight(fHeight)
        for k,v in pairs(self) do
            if type(v) == "table" and v.SetFont then
                v:SetFont(fontName, fontHeight, fontFlags)
                v:SetWidth(v.id == 1 and size1 or size2)
            end
        end
    end,
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function Constructor()
    local widget = {}
    widget.type  = Type

    local frame = CreateFrame("Frame")
    frame:SetWidth(76)
    frame:SetHeight(20)
    widget.frame = frame

    local timer_lH = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_lH:SetPoint("LEFT",frame,"LEFT",0,0)
    timer_lH:SetWordWrap(false)
    timer_lH:SetJustifyH("CENTER")
    timer_lH:SetJustifyV("MIDDLE")
    timer_lH.id = 1
    widget.timer_lH = timer_lH

    local timer_t1 = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_t1:SetPoint("LEFT",timer_lH,"RIGHT",0,0)
    timer_t1:SetWordWrap(false)
    timer_t1:SetJustifyH("CENTER")
    timer_t1:SetJustifyV("MIDDLE")
    timer_t1:SetText(":")
    timer_t1.id = 2
    widget.timer_t1 = timer_t1

    local timer_lM = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_lM:SetPoint("LEFT",timer_t1,"RIGHT",0,0)
    timer_lM:SetWordWrap(false)
    timer_lM:SetJustifyH("CENTER")
    timer_lM:SetJustifyV("MIDDLE")
    timer_lM.id = 1
    widget.timer_lM = timer_lM

    local timer_t2 = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_t2:SetPoint("LEFT",timer_lM,"RIGHT",0,0)
    timer_t2:SetWordWrap(false)
    timer_t2:SetJustifyH("CENTER")
    timer_t2:SetJustifyV("MIDDLE")
    timer_t2:SetText(":")
    timer_t2.id = 2
    widget.timer_t2 = timer_t2

    local timer_lS = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_lS:SetPoint("LEFT",timer_t2,"RIGHT",0,0)
    timer_lS:SetWordWrap(false)
    timer_lS:SetJustifyH("CENTER")
    timer_lS:SetJustifyV("MIDDLE")
    timer_lS.id = 1
    widget.timer_lS = timer_lS

    widget.TimerUpdate = function()
        if not widget.isRunning then return end
        if widget.endTime <= widget.timeFunc() then
            widget:FinishTimer()
            widget:StopTimer()
            return
        end
        Timer_SetData(widget)
        C_TimerAfter(TIMER_INTERVAL, widget.TimerUpdate)
    end

    -- methodes
    for method, func in pairs(methods) do
        widget[method] = func
    end

    return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)
