local AddonName, GGL = ...

local AceGUI = LibStub("AceGUI-3.0")
local Templates = GGL.Templates
local L = GGL.Locale

-- lua
local pairs, assert, type = pairs, assert, type
local tblsort, tblremove = table.sort, table.remove
local xpcall = xpcall
local time, date = time, date
local format = string.format

-- wow

local MainWindow = {}
GGL.GUI.MainWindow = MainWindow

local GetMoneyText = GGL.GetMoneyText
local CreateLine = GGL.GUI.CreateLine

local function OnMapChange(sender, mapID, mapName, mapTypeFlag)
    if not mapID then return end
    local frame = MainWindow.frame
    frame:UpdateZone(mapName, mapTypeFlag)
    if frame.ZoneFilter and not frame.ZoneFilter:IsValidZone(mapID) then
        frame.contentContainer.zoneWarningFrame.warningText = format(L["Zone '%s' is forbidden."], mapName or "")
        frame.contentContainer.zoneWarningFrame:Show()
    else
        frame.contentContainer.zoneWarningFrame:Hide()
    end
end

local function OnInstancesResetUpdate(sender, instanceLocks, nextReleaseStartTime, nextReleaseEndTime)
    MainWindow.frame:UpdateReset(instanceLocks, nextReleaseStartTime, nextReleaseEndTime)
end

function MainWindow:Show()
    self:CreateFrame()

    self.frame:Show()

    OnMapChange(nil, GGL:GetZoneData())

    return self.frame
end

-- TODO: remove
function MainWindow:LoadSession(session, options)
    self:CreateFrame()

    self.frame:Show()

    session = session or Templates:New("Session", GGL.sessionDb.Session)

    if options and not session:DataIsSet() then
        session:SetData((options.duration or 5)*60, options.set or "Base")
    end
    session:AddIntoMainFrame(true)

    session:Start()
end

-- ##################
-- Window Creation
-- ##################
local WIDTH_V, HEIGHT_V = 370, 212
local WIDTH_M, HEIGHT_M = 100, 100
local MAX_ITEM_ENTRYS = 50
local TIME_FORMAT = "%H:%M"
local TT_INFO_ENTRY = "|cFFCFCFCF%s:|r %s"

local BACKDROP = {
    bgFile = [[Interface\Tooltips\UI-Tooltip-Background]],
    edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
    edgeSize = 2,
    insets = { left = 1, right = 1, top = 1, bottom = 1 }
}

local FRAME_FUNCTIONS = {
    ["Clear"] = function(self)
        self.itemContainer:ReleaseChildren()

        -- reset gold
        self:SetValueGold(0)
        self:SetRawGold(0)
        self:SetTotalGold(0)
    end,

    ["SetMaxItemEntrys"] = function(self, num)
        self.maxItemEntrys = num
        self.itemContainer:ReleaseChildren()
    end,

    ["AddItem"] = function(self, itemID, sourceItemID, quantity, moneyValue, mainSource, subSource)
        if not itemID then return end
        local children = self.itemContainer.children
        local itemRow

        -- clear first item if we are over cap
        if #children+1 > self.maxItemEntrys then
            local content = self.itemContainer.content
            itemRow = tblremove(children, 1)
            local frame = itemRow.frame

            frame:ClearAllPoints()
            children[#children].frame:SetPoint("BOTTOMLEFT", frame, "TOPLEFT")
            frame:SetPoint("BOTTOMLEFT", content)
            frame:SetPoint("RIGHT", content)

            children[#children+1] = itemRow
        end

        if not itemRow then
            itemRow = AceGUI:Create("GoldGoblinLootGUIWidget-ItemLine")
            itemRow:SetFullWidth(true)
            self.itemContainer:AddChild(itemRow)
            itemRow:SetToolTipPoint("BOTTOM_X", self.frame)
        end

        itemRow:SetItem(itemID, sourceItemID, quantity, moneyValue, mainSource, subSource)
        self.itemContainer:SetScroll(1000)

        self.itemContainer.frame:Show()
    end,

    ["SetValueGold"] = function(self, copper, add)
        self.valueGoldCount = add and (self.valueGoldCount or 0) + copper or copper
        self.contentContainer.valueGold:SetText(GetMoneyText(self.valueGoldCount))
        self:SetTotalGold((self.rawGoldCount or 0) + (self.valueGoldCount or 0))
    end,

    ["SetRawGold"] = function(self, copper, add)
        if not copper or copper < 0 then return end
        self.rawGoldCount = add and ((self.rawGoldCount or 0) + copper) or copper
        self.contentContainer.rawGold:SetText(GetMoneyText(self.rawGoldCount))
        self:SetTotalGold((self.rawGoldCount or 0) + (self.valueGoldCount or 0))
    end,

    ["SetTotalGold"] = function(self, copper, add)
        self.sumGoldCount = add and (self.sumGoldCount or 0) + copper or copper
        self.contentContainer.sumGold:SetText(GetMoneyText(self.sumGoldCount))
    end,

    ["UpdateReset"] = function(self, instanceLocks, nextReleaseStartTime, nextReleaseEndTime)
        local frame = self.contentContainer
        if not instanceLocks or not nextReleaseEndTime then
            frame.resetLockout:SetFormattedText(L["Instance lockout: %d/10"], 0)
            frame.instanceResetTimer:StopTimer()
            frame.resetLockoutFrame.instanceLocks = nil
        else
            frame.resetLockout:SetFormattedText(L["Instance lockout: %d/10"], #instanceLocks)
            frame.instanceResetTimer:SetTime(nextReleaseStartTime, nextReleaseEndTime)
            frame.instanceResetTimer:StartTimer()
            frame.resetLockoutFrame.instanceLocks = instanceLocks
        end
        -- refresh tooltip
        if frame.resetLockoutFrame.isOnEnter then
            frame.resetLockoutFrame:GetScript("OnEnter")(frame.resetLockoutFrame)
        end
    end,

    ["UpdateZone"] = function(self, mapName, mapTypeFlag)
        local frame = self.contentContainer

        if mapTypeFlag == GGL.MAP_TYPE_FLAGS.RAID then
            frame.zoneIcon:SetAtlas("Raid")
        elseif mapTypeFlag == GGL.MAP_TYPE_FLAGS.DUNGEON then
            frame.zoneIcon:SetAtlas("Dungeon")
        else
            frame.zoneIcon:SetAtlas("Vehicle-Carriage")
        end

        frame.zoneText:SetText(mapName)
    end,

    ["RefreshMap"] = function(self)
        OnMapChange(nil, GGL:GetZoneData())
    end,
}

local function ZoneWarning_OnEnter(self)
    -- contentContainer.zoneWarningFrame
    if self.warningText then
        local tooltip = _G.GameTooltip
        tooltip:SetOwner(self, "ANCHOR_NONE")
        tooltip:ClearAllPoints()
        tooltip:SetPoint("BOTTOMLEFT", self.ttFrame, "BOTTOMRIGHT", 2, 0)
        tooltip:AddLine(L["Warning:"])
        tooltip:AddLine(self.warningText)
        tooltip:Show()
    end
end

local function ResetLockout_OnEnter(self)
    local tooltip = _G.GameTooltip
    tooltip:ClearLines()
    GGL.GUI:SetToolTipPoint(tooltip, self.ttFrame, "BOTTOM_X")
    tooltip:AddLine(format(L["Instance lockout: %d/10"], self.instanceLocks and #self.instanceLocks or 0))
    if self.instanceLocks then
        for i = 1, #self.instanceLocks do
            tooltip:AddDoubleLine(date(TIME_FORMAT, self.instanceLocks[i].endTime), self.instanceLocks[i].zone)
        end
    end
    tooltip:AddLine(format(TT_INFO_ENTRY, L["Alt + Click"], L["Reset Instances"]))
    tooltip:Show()
    self.isOnEnter = true
end

local function ResetLockout_OnLeave(self)
    _G.GameTooltip:Hide()
    self.isOnEnter = false
end

local function ResetLockout_OnClick(self, button)
    if IsAltKeyDown() then
        ResetInstances()
    end
end

function MainWindow:CreateFrame()
    if not self.frame then
        AceGUI = LibStub("AceGUI-3.0")

        local window = AceGUI:Create("Window")
        --window:SetLayout("Flow")
        window:SetTitle("GoldGoblinLoot")

        local frame = window.frame
        local content = window.content

        -- some frame settings
        frame:SetMinResize(WIDTH_V,HEIGHT_V)
        frame:SetMaxResize(WIDTH_V+WIDTH_M,HEIGHT_V+HEIGHT_M)
        frame:SetClampedToScreen(true)
        frame:SetToplevel(false)
        frame:SetFrameStrata("MEDIUM")
        --window settings
        window:SetWidth(WIDTH_V)
        window:SetHeight(HEIGHT_V)
        window.maxItemEntrys = MAX_ITEM_ENTRYS
        -- remove resizer lines
        window.line1:Hide()
        window.line2:Hide()

        --itemContainerBG
        local itemContainerBG = GGL.GUI:CreateBorderedFrame(frame)

        -- itemContainer
        local itemContainer = AceGUI:Create("ScrollFrame")
        itemContainer.frame:SetParent(content) -- fix for a bug with skinning
        itemContainer.frame:SetPoint("TOPLEFT", content)
        itemContainer:SetLayout("GGL-Down2Up")
        --itemContainer
        window.itemContainer = itemContainer

        itemContainerBG:SetPoint("TOPLEFT", itemContainer.scrollframe, -5, 5)
        itemContainerBG:SetPoint("BOTTOMRIGHT", itemContainer.scrollframe, 5, -5)

        -- contentContainer
        local contentContainerF = GGL.GUI:CreateBorderedFrame(frame)

        local contentContainer = CreateFrame("Frame", nil, contentContainerF)
        contentContainer:SetPoint("BOTTOMLEFT", content, 0, 0)
        contentContainer:SetPoint("BOTTOMRIGHT", content, 0, 0)
        contentContainer:SetHeight(87)
        window.contentContainer = contentContainer

        contentContainerF:SetPoint("TOPLEFT", contentContainer, -5, 5)
        contentContainerF:SetPoint("BOTTOMRIGHT", contentContainer, 5, -5)

        itemContainer.frame:SetPoint("BOTTOMRIGHT", contentContainer, "TOPRIGHT", 0, 8)

        -- content:Timer
        local timer = AceGUI:Create("GoldGoblinLootGUIWidget-Timer")
        timer.frame:SetParent(contentContainer)
        timer:SetPoint("TOPRIGHT", contentContainer, "TOPRIGHT", 0, 0)
        timer:SetDuration(0)
        contentContainer.timer = timer

        -- content: ZoneIcon
        local zoneIcon = contentContainer:CreateTexture(nil, "BACKGROUND")
        zoneIcon:SetPoint("TOPLEFT", contentContainer)
        zoneIcon:SetWidth(timer.frame:GetHeight())
        zoneIcon:SetHeight(timer.frame:GetHeight())
        zoneIcon:SetAtlas("Vehicle-Carriage")
        contentContainer.zoneIcon = zoneIcon

        -- content: Zone
        local zoneText = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        zoneText:SetPoint("TOPLEFT", zoneIcon, "TOPRIGHT", 2, 0)
        zoneText:SetPoint("BOTTOMRIGHT", timer.frame, "BOTTOMLEFT")
        zoneText:SetJustifyH("LEFT")
        zoneText:SetJustifyV("CENTER")
        contentContainer.zoneText = zoneText

        local zoneWarningFrame = CreateFrame("Frame", nil, contentContainer)
        zoneWarningFrame:SetPoint("RIGHT", zoneText, -5, 0)
        zoneWarningFrame:SetWidth(timer.frame:GetHeight())
        zoneWarningFrame:SetHeight(timer.frame:GetHeight())
        zoneWarningFrame:SetScript("OnEnter", ZoneWarning_OnEnter)
        zoneWarningFrame:SetScript("OnLeave", ResetLockout_OnLeave)
        zoneWarningFrame.ttFrame = frame
        zoneWarningFrame:Hide()
        contentContainer.zoneWarningFrame = zoneWarningFrame

        local zoneWarning = zoneWarningFrame:CreateTexture(nil, "BACKGROUND")
        zoneWarning:SetAllPoints(zoneWarningFrame)
        zoneWarning:SetAtlas(GGL.ATLAS_TEXTURE.WARNING)
        zoneWarningFrame.texture = zoneWarning

        -- content: RawGold
        local rawGold = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        rawGold:SetPoint("TOPLEFT", zoneIcon, "BOTTOMLEFT", 0, -2)
        rawGold:SetPoint("LEFT", contentContainer)
        rawGold:SetPoint("RIGHT", contentContainer, "CENTER", 0, 0)
        rawGold:SetHeight(23)
        rawGold:SetJustifyH("CENTER")
        rawGold:SetJustifyV("BOTTOM")
        contentContainer.rawGold = rawGold

        local rawGoldTitle = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        rawGoldTitle:SetAllPoints(rawGold)
        rawGoldTitle:SetJustifyH("LEFT")
        rawGoldTitle:SetJustifyV("TOP")
        rawGoldTitle:SetText(L["Raw gold"]..":")
        contentContainer.rawGoldTitle = rawGoldTitle

        -- content: ValueGold
        local valueGold = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        valueGold:SetPoint("TOPLEFT", rawGoldTitle, "TOPRIGHT", 2, 0)
        valueGold:SetPoint("BOTTOMLEFT", rawGoldTitle, "BOTTOMRIGHT")
        valueGold:SetPoint("RIGHT", contentContainer)
        valueGold:SetJustifyH("CENTER")
        valueGold:SetJustifyV("BOTTOM")
        contentContainer.valueGold = valueGold

        local valueGoldTitle = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        valueGoldTitle:SetAllPoints(valueGold)
        valueGoldTitle:SetJustifyH("LEFT")
        valueGoldTitle:SetJustifyV("TOP")
        valueGoldTitle:SetText(L["Item value gold:"])
        contentContainer.valueGoldTitle = valueGoldTitle

        -- content: ValueGold
        local sumGold = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        sumGold:SetPoint("TOPLEFT", rawGold, "BOTTOMLEFT", 0, -3)
        sumGold:SetPoint("RIGHT", contentContainer)
        sumGold:SetHeight(rawGold:GetHeight())
        sumGold:SetJustifyH("CENTER")
        sumGold:SetJustifyV("BOTTOM")
        contentContainer.sumGold = sumGold

        local sumGoldTitle = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        sumGoldTitle:SetAllPoints(sumGold)
        sumGoldTitle:SetJustifyH("LEFT")
        sumGoldTitle:SetJustifyV("TOP")
        sumGoldTitle:SetText(L["Total gold:"])
        contentContainer.sumGoldTitle = sumGoldTitle

        -- content: instanceReset

        local instanceResetTimer = AceGUI:Create("GoldGoblinLootGUIWidget-Timer")
        instanceResetTimer.frame:SetParent(contentContainer)
        instanceResetTimer:SetPoint("TOPRIGHT", sumGold, "BOTTOMRIGHT", -2, -2)
        instanceResetTimer:SetDuration(0)
        instanceResetTimer:SetSize("small")
        instanceResetTimer.frame:SetHeight(14)
        contentContainer.instanceResetTimer = instanceResetTimer

        local resetLockoutRelease = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        resetLockoutRelease:SetPoint("TOPRIGHT", instanceResetTimer.frame, "TOPLEFT", 0, 0)
        resetLockoutRelease:SetHeight(instanceResetTimer.frame:GetHeight())
        resetLockoutRelease:SetJustifyH("LEFT")
        resetLockoutRelease:SetJustifyV("MIDDLE")
        resetLockoutRelease:SetText(L["Next Instance lockout release:"])
        resetLockoutRelease:SetWidth(resetLockoutRelease:GetStringWidth()+3)
        contentContainer.resetLockoutRelease = resetLockoutRelease

        local resetLockout = contentContainer:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall")
        resetLockout:SetPoint("TOPLEFT", sumGold, "BOTTOMLEFT", 0, -2)
        resetLockout:SetHeight(instanceResetTimer.frame:GetHeight())
        resetLockout:SetJustifyH("LEFT")
        resetLockout:SetJustifyV("MIDDLE")
        resetLockout:SetFormattedText(L["Instance lockout: %d/10"], 999)
        resetLockout:SetWidth(resetLockout:GetStringWidth()+20)
        contentContainer.resetLockout = resetLockout

        local resetLockoutFrame = CreateFrame("BUTTON", nil, contentContainer)
        resetLockoutFrame:SetPoint("TOPLEFT", resetLockout)
        resetLockoutFrame:SetPoint("BOTTOMRIGHT", instanceResetTimer.frame)
        resetLockoutFrame:SetScript("OnEnter", ResetLockout_OnEnter)
        resetLockoutFrame:SetScript("OnLeave", ResetLockout_OnLeave)
        resetLockoutFrame:SetScript("OnClick", ResetLockout_OnClick)
        resetLockoutFrame.ttFrame = frame
        contentContainer.resetLockoutFrame = resetLockoutFrame


        -- lines
        local lines = {
            -- Timer left
            CreateLine(contentContainer, true, {"TOPRIGHT",timer.frame,"TOPLEFT",-2,0}, {"BOTTOMRIGHT",timer.frame,"BOTTOMLEFT"}, nil, nil),
            -- Timer bottom
            CreateLine(contentContainer, true, {"LEFT"}, {"TOPRIGHT",timer.frame,"BOTTOMRIGHT"}, nil, nil),
            -- raw/value gold center
            CreateLine(contentContainer, true, {"TOP", rawGold, "TOPRIGHT", 0, 1}, {"BOTTOM", rawGold, "BOTTOMRIGHT", 0, -2}, nil, nil),
            -- raw/value gold bottom
            CreateLine(contentContainer, true, {"LEFT", rawGold, "BOTTOMLEFT", 0, -2}, {"RIGHT"}, nil, nil),
            -- total gold bottom
            CreateLine(contentContainer, true, {"LEFT", sumGold, "BOTTOMLEFT", 0, -2}, {"RIGHT"}, nil, nil),
            -- resetLockout middle
            CreateLine(contentContainer, true, {"TOPRIGHT", resetLockoutRelease, "TOPLEFT", -1, 0}, {"BOTTOMRIGHT", resetLockoutRelease, "BOTTOMLEFT", -1, 0}, nil, nil),
            -- resetLockout bottom
            --CreateLine(contentContainer, true, {"TOPLEFT", resetLockout, "BOTTOMLEFT", 0, 0}, {"RIGHT"}, nil, nil),
        }

        -- register functions
        for k,v in pairs(FRAME_FUNCTIONS) do
            window[k] = v
        end
        FRAME_FUNCTIONS = nil
        self.frame = window

        GGL:RegisterCallbackEvent("OnMapChange", OnMapChange)
        GGL:RegisterCallbackEvent("OnInstancesResetUpdate", OnInstancesResetUpdate)

        GGL:ZONE_CHANGED_NEW_AREA()
        GGL:UpdateInstanceResets()

        GGL.GUI:AddFrameControl("MainWindow", frame, true, true)

        window:Hide()
    end
    return self.frame
end

function MainWindow:AddIntoMiniMapButton()
    GGL:RegisterMMFunction(self, self.Show)
end

function MainWindow:RemoveFromMiniMapButton()
    GGL:UnregisterMMFunction(self.Show)
end

-- ##################
-- Layout
-- ##################
local function safecall(func, ...)
	if func then
		return xpcall(func, error, ...)
	end
end

AceGUI:RegisterLayout("GGL-Down2Up",
    function(content, children)
        local height = 0
        local width = content.width or content:GetWidth() or 0
        for i = 1, #children do
            local child = children[i]

            local frame = child.frame
            frame:ClearAllPoints()
            frame:Show()
            if i == 1 then
                frame:SetPoint("BOTTOMLEFT", content)
            else
                children[i-1].frame:SetPoint("BOTTOMLEFT", frame, "TOPLEFT")
                frame:SetPoint("BOTTOMLEFT", content)
            end

            if child.width == "fill" then
                child:SetWidth(width)
                frame:SetPoint("RIGHT", content)

                if child.DoLayout then
                    child:DoLayout()
                end
            elseif child.width == "relative" then
                child:SetWidth(width * child.relWidth)

                if child.DoLayout then
                    child:DoLayout()
                end
            end

            height = height + (frame.height or frame:GetHeight() or 0)
        end
        safecall(content.obj.LayoutFinished, content.obj, nil, height)
    end)