local AddonName, GGL = ...

local AceGUI = LibStub("AceGUI-3.0")
local Templates = GGL.Templates
local L = GGL.Locale

local TopItems = {}
GGL.GUI.TopItems = TopItems

local pairs = pairs
local tblsort = table.sort

local GetPlayerTemplate = Templates.Sync.GetPlayerFromSync

-- ####################
-- TopItems frame

-- EVENTS
-- ####################
local WIDTH_V, HEIGHT_V = 300, 150
local WIDTH_M, HEIGHT_M = 200, 1000

local NUM_MAX_ITEMS = GGL.MAX_TOP_ITEMS

local function SorterFunction(a, b)
    return a.value > b.value
end

local FRAME_FUNCTIONS = {
    ["Clear"] = function(self)
        self:ClearItemRows(true)

        self:SetSession()
    end,

    --- Sets the linked session to the frame
    -- @param       sessionTemplate         <sessionTemplate>
    ["SetSession"] = function(self, sessionTemplate)
        self.session = sessionTemplate
    end,

    --- Clears all items from the list
    -- @param       clearItems      <bool> Clear also all items from the list
    ["ClearItemRows"] = function(self, clearItems)
        self.itemContainer:ReleaseChildren()
        if clearItems then
            wipe(self.ItemList)
            wipe(self.ItemListByID)
            self:UpdateLowestItemValueDefault()
        end
        self:UpdateItemList()
    end,

    --- Sets the lowest item value of the list
    -- this defaults to the value set in options
    -- @param       lowestItemValue     <number>
    ["UpdateLowestItemValueDefault"] = function(self, lowestItemValue)
        self.lowestItemValueDefault = lowestItemValue or GGL.db.Settings.TopItems.minGoldValue
    end,

    --- Adds a new item
    -- Checks first if its valid
    ["AddItem"] = function(self, playerTemplate, itemID, ammount, copperValue, zoneID, sourceID, mainSourceFlag, subSourceFlag)
        -- item not worth it
        if copperValue < self.lowestItemValueDefault then return end

        local workingTable = self.ItemListByID[itemID]
        local refreshFullList = false
        local playerKey = self.session:GetPlayerMergeID(playerTemplate) or playerTemplate.syncID

        if not workingTable then
            -- new item !!!
            self.ItemListByID[itemID] = {
                itemID = itemID,
                value = copperValue,    -- this is fix
                ammount = ammount,      -- full ammount from all palyers
                players = {},
                playerSortedList = {},  -- sorted list of all players
            }
            self.ItemList[#self.ItemList+1] = self.ItemListByID[itemID]

            -- set workingTable
            workingTable = self.ItemListByID[itemID]
            refreshFullList = true
        else
            workingTable.ammount = workingTable.ammount + ammount
        end

        -- refresh player data
        if not workingTable.players[playerKey] then
            workingTable.players[playerKey] = {
                playerKey = playerKey,
                playerTemplate = playerTemplate,
                ammount = ammount,
                dropData = {},
            }
            workingTable.playerSortedList[ #workingTable.playerSortedList + 1 ] = workingTable.players[playerKey]
        else
            workingTable.players[playerKey].ammount = workingTable.players[playerKey].ammount + ammount
        end

        -- add drop data that is needed for the tooltip later
        workingTable.players[playerKey].dropData[ #workingTable.players[playerKey].dropData + 1] = {
            ammount,
            zoneID,
            sourceID,
            mainSourceFlag,
            subSourceFlag
        }

        -- for new items we need a full list update
        if refreshFullList then
            self:UpdateItemList()
        elseif workingTable.row then
            workingTable.row:UpdateItem()
            -- update players
            workingTable.row:AddPlayer(workingTable.players[playerKey])
        end
    end,

    ["UpdateMergeList"] = function(self)
        if not self.session then return end

        for index, itemData in ipairs(self.ItemList) do
            if itemData.row then
                for index, playerData in ipairs(itemData.playerSortedList) do
                    -- get active player for a key
                    local activePlayerTemplate = self.session:GetActivePlayerForMergeID(playerData.playerKey) or playerData.playerTemplate
                    -- get playerKey
                    local playerKey = self.session:GetPlayerMergeID(activePlayerTemplate) or activePlayerTemplate.syncID
                    -- playerKey changed
                    if playerData.playerKey ~= playerKey then
                        if itemData.players[playerKey] then
                            for index, item in ipairs(itemData.players[playerData.playerKey].dropData) do
                                itemData.players[playerKey].dropData[ #itemData.players[playerKey].dropData+1 ] = item
                            end

                            itemData.playerSortedList[index] = itemData.players[playerKey]
                            playerData = itemData.players[playerKey]
                        else
                            -- update playerKey list
                            itemData.players[playerKey] = playerData
                            -- update playerKey
                            playerData.playerKey = playerKey
                        end

                        itemData.row:RemovePlayer(itemData.players[playerData.playerKey])
                        itemData.players[playerData.playerKey] = nil
                    end

                    -- playerTemplate changed
                    if playerData.playerTemplate ~= activePlayerTemplate then
                        -- update playerTemplate
                        playerData.playerTemplate = activePlayerTemplate
                    end

                    -- update player
                    itemData.row:AddPlayer(playerData)

                    itemData.row:UpdateItem()
                end
            end
        end
    end,

    --- Updates the itemList
    ["UpdateItemList"] = function(self)
        -- sort all entrys by value
        tblsort(self.ItemList, SorterFunction)

        -- pause Layout for now
        self.itemContainer:PauseLayout()

        -- set new row sorting
        for index, itemData in ipairs(self.ItemList) do
            if index <= GGL.db.Settings.TopItems.numTopItems then
                if not itemData.row then
                    local row = AceGUI:Create("GoldGoblinLootGUIWidget-TopItem")
                    row:SetFullWidth(true)
                    row:SetItemData(itemData)
                    row:SetToolTipPoint("BOTTOM_X", self.frame)
                    self.itemContainer:AddChild(row)
                    itemData.row = row

                    -- update playerdata
                    for index, playerData in ipairs(itemData.playerSortedList) do
                        row:AddPlayer(playerData)
                    end
                else
                    itemData.row:UpdateItem()
                end
                -- set new position for the frame
                self.itemContainer.children[index] = itemData.row
                itemData.row:SetListPosition(index)
            else
                if itemData.row then
                    itemData.row:Release()
                    itemData.row = nil
                end
                self.itemContainer.children[index] = nil
                if index > NUM_MAX_ITEMS then
                    -- remove item from list
                    self.ItemListByID[itemData.itemID] = nil
                    self.ItemList[index] = nil
                end
            end
        end

        -- resume layout und run it
        self.itemContainer:ResumeLayout()
        self.itemContainer:DoLayout()
    end,

    --- Sets the inviteState for a player
    -- @param       player      <playerTemplate>
    -- @param       state       <GGL.INVITE_KEY_LIST>
    ["SetPlayerInviteState"] = function(self, player, state)
        if not player then return end

        for index, itemData in ipairs(self.ItemList) do
            if itemData.row then
                if state == GGL.INVITE_KEY_LIST.ACCEPTED then
                    itemData.row:RegisterPlayerEvents(self.session:GetPlayerMergeID(player) or player.syncID, player)
                else
                    itemData.row:UnregisterPlayerEvents(player)
                end
            end
        end
    end,

    --- Unregisters all player events like name changes
    ["UnregisterAllPlayerEvents"] = function(self)
        for index, itemData in ipairs(self.ItemList) do
            if itemData.row then
                itemData.row:UnregisterAllPlayerEvents()
            end
        end
    end,
}

local function TopItems_OnDbUpdate(event, db, key, value)
    if not TopItems.frame then return end
    local frame = TopItems.frame
    frame:UpdateItemList()
    -- frame:UpdateLowestItemValueDefault()
end

function TopItems:CreateFrame()
    if not self.frame then
        local window = AceGUI:Create("Window")
        window:SetTitle("GoldGoblinLoot - TopItems")

        local frame = window.frame
        local content = window.content

        -- some frame settings
        frame:SetMinResize(WIDTH_V,HEIGHT_V)
        frame:SetMaxResize(WIDTH_V+WIDTH_M,HEIGHT_V+HEIGHT_M)
        frame:SetClampedToScreen(true)
        frame:SetToplevel(false)
        frame:SetFrameStrata("MEDIUM")
        --window settings
        window:SetWidth(WIDTH_V)
        window:SetHeight(HEIGHT_V+120)
        -- remove resizer lines
        window.line1:Hide()
        window.line2:Hide()

        --itemContainerBG
        local itemContainerBG = GGL.GUI:CreateBorderedFrame(content)
        -- itemContainer
        local itemContainer = AceGUI:Create("ScrollFrame")
        itemContainer:SetLayout("List")
        itemContainer:SetFullWidth(true)
        itemContainer:SetFullHeight(true)
        window.itemContainer = itemContainer
        window:AddChild(itemContainer)
        -- itemContainer / bg
        itemContainerBG:SetPoint("TOPLEFT", itemContainer.scrollframe, -5, 5)
        itemContainerBG:SetPoint("BOTTOMRIGHT", itemContainer.scrollframe, 5, -5)

        -- register functions
        for k,v in pairs(FRAME_FUNCTIONS) do
            window[k] = v
        end
        FRAME_FUNCTIONS = nil

        GGL:AddCallbackHandler(window)

        window.ItemList = {}
        window.ItemListByID = {}

        -- cb
        GGL.Options:RegisterCallbackEvent("DbOnSet-Settings-TopItems-numTopItems", TopItems_OnDbUpdate)
        --GGL.Options:RegisterCallbackEvent("DbOnSet-Settings-TopItems-minGoldValue", TopItems_OnDbUpdate)

        GGL.GUI:AddFrameControl("TopItems", frame, true, true)

        self.frame = window
    end
    return self.frame
end

function TopItems:Show()
    TopItems:CreateFrame():Show()
    return self.frame
end

function TopItems:AddIntoMiniMapButton()
    GGL:RegisterMMFunction(self, self.Show)
end

function TopItems:RemoveFromMiniMapButton()
    GGL:UnregisterMMFunction(self.Show)
end
